from typing import Any, Literal, NotRequired, TypedDict

from mipac.types.ads import IPartialAd

ISensitiveMediaDetectionSentivity = Literal["medium", "low", "high", "veryLow", "veryHigh"]
ISensitiveMediaDetection = Literal["none", "all", "local", "remote"]


class IFeatures(TypedDict):
    registration: bool
    email_required_for_signup: bool
    hcaptcha: bool
    recaptcha: bool
    turnstile: bool
    object_storage: bool
    service_worker: bool
    miauth: bool


class IPolicies(TypedDict):
    gtl_available: bool
    ltl_available: bool
    can_public_note: bool
    can_edit_note: bool
    can_invite: bool
    invite_limit: int
    invite_limit_cycle: int
    invite_expiration_time: int
    can_manage_custom_emojis: bool
    can_search_notes: bool
    can_use_translator: bool
    can_hide_ads: bool
    drive_capacity_mb: int
    always_mark_nsfw: bool
    pin_limit: int
    antenna_limit: int
    word_mute_limit: int
    webhook_limit: int
    clip_limit: int
    note_each_clips_limit: int
    user_list_limit: int
    user_each_user_lists_limit: int
    rate_limit_factor: int


class IPartialMeta(TypedDict):
    maintainer_name: str | None
    maintainer_email: str | None
    version: str
    name: None
    short_name: None
    uri: str
    description: None
    langs: list[str]
    tos_url: str
    repository_url: str
    feedback_url: str
    disable_registration: bool
    email_required_for_signup: bool
    enable_hcaptcha: bool
    hcaptcha_site_key: None
    enable_recaptcha: bool
    recaptcha_site_key: str
    enable_turnstile: bool
    turnstile_site_key: str
    sw_publickey: None
    theme_color: str
    mascot_image_url: str
    banner_url: str | None
    info_image_url: str | None
    server_error_image_url: str | None
    not_found_image_url: str | None
    icon_url: str | None
    background_image_url: str | None
    logo_image_url: str | None
    max_note_text_length: int
    default_light_theme: str | None
    default_dark_theme: str | None
    ads: list[IPartialAd]
    enable_email: bool
    enable_service_worker: bool
    translator_available: bool
    inquiry_url: str | None
    server_rules: list[str]
    policies: IPolicies
    media_proxy: str


class IMetaDetailedOnly(TypedDict):
    features: IFeatures
    proxy_account_name: str
    require_setup: bool
    cache_remote_files: bool
    cache_remote_sensitive_files: bool


class IMeta(IPartialMeta, TypedDict):
    ...


class IAdminMeta(TypedDict):  # IMetaに含まれる物が多くあるけど、ない場合もあるので別にする
    maintainer_name: str | None
    maintainer_email: str | None
    version: str
    name: str | None
    short_name: str | None
    uri: str
    description: str | None
    langs: list[str]
    tos_url: str
    repository_url: str
    feedback_url: str
    disable_registration: bool
    email_required_for_signup: bool
    enable_hcaptcha: bool
    hcaptcha_site_key: str | None
    enable_recaptcha: bool
    recaptcha_site_key: str
    enable_turnstile: bool
    turnstile_site_key: str
    sw_publickey: str | None
    theme_color: str
    mascot_image_url: str
    banner_url: str | None
    server_error_image_url: str | None
    not_found_image_url: str | None
    info_image_url: str | None
    icon_url: str | None
    appint_icon_url: str | None
    appint_icon_url: str | None
    background_image_url: str | None
    logo_image_url: str | None
    default_light_theme: str | None
    default_dark_theme: str | None
    enable_email: bool
    enable_service_worker: bool
    translator_available: bool
    cache_remote_files: bool
    cache_remote_sensitive_files: bool
    pinned_users: list[str]
    hidden_tags: list[str]
    blocked_hosts: list[str]
    sensitive_words: list[str]
    preserved_usernames: list[str]
    hcaptcha_secret_key: str | None
    recaptcha_secret_key: str
    turnstile_secret_key: str
    sensitive_media_detection: ISensitiveMediaDetection
    sensitive_media_detection_sensitivity: ISensitiveMediaDetectionSentivity
    set_sensitive_flag_automatically: bool
    enable_sensitive_media_detection_for_videos: bool
    proxy_account_id: str
    summaly_proxy: str | None
    email: str
    smtp_secure: bool
    smtp_host: str
    smtp_port: int
    smtp_user: str
    smtp_pass: str
    sw_private_key: str | None
    use_object_storage: bool
    object_storage_base_url: str
    object_storage_bucket: str
    object_storage_prefix: str
    object_storage_endpoint: str
    object_storage_region: str
    object_storage_port: str | None
    object_storage_access_key: str
    object_storage_secret_key: str
    object_storage_use_s_s_l: bool
    object_storage_use_proxy: bool
    object_storage_set_public_read: bool
    object_storage_sint_force_path_style: bool
    deepl_auth_key: str | None
    deepl_is_pro: bool
    enable_ip_logging: bool
    enable_active_email_validation: bool
    enable_charts_for_remote_user: bool
    enable_charts_for_federated_instances: bool
    enable_server_machine_stats: bool
    enable_identicon_generation: bool
    policies: IPolicies
    manifest_json_override: dict[str, Any]


class IUpdateMetaBody(TypedDict, total=False):
    announcements: list
    disable_registration: bool
    disable_local_timeline: bool
    disable_global_timeline: bool
    enable_emoji_reaction: bool
    use_star_for_reaction_fallback: bool
    pinned_users: list[str]
    hidden_tags: list[str]
    blocked_hosts: list[str]
    mascot_image_url: str | None
    banner_url: str | None
    error_image_url: str | None
    icon_url: str | None
    name: str | None
    description: str | None
    max_note_text_length: int
    local_drive_capacity_mb: int
    remote_drive_capacity_mb: int
    cache_remote_files: bool
    proxy_remote_files: bool
    enable_recaptcha: bool
    recaptcha_site_key: str
    recaptcha_secret_key: str
    enable_turnstile: bool
    turnstile_site_key: str
    turnstile_secret_key: str
    proxy_account_id: str | None
    proxy_account: str
    maintainer_name: str | None
    maintainer_email: str | None
    langs: str
    summaly_proxy: str | None
    enable_twitter_integration: bool
    twitter_consumer_key: str | None
    twitter_consumer_secret: str | None
    enable_github_integration: bool
    github_client_id: str | None
    github_client_secret: str | None
    enable_discord_integration: bool
    discord_client_id: str | None
    discord_client_secret: str | None
    enable_email: bool
    email: str
    smtp_secure: bool
    smtp_host: str | None
    smtp_port: int | None
    smtp_user: str | None
    smtp_pass: str | None
    enable_service_worker: bool
    sw_public_key: str
    sw_private_key: str
    tos_url: str | None
    tos_text_url: str
    repository_url: str
    feedback_url: str
    use_object_storage: bool
    object_storage_base_url: str | None
    object_storage_bucket: str | None
    object_storage_prefix: str | None
    object_storage_endpoint: str | None
    object_storage_region: str | None
    object_storage_port: int | None
    object_storage_access_key: str | None
    object_storage_secret_key: str | None
    object_storage_use_s_s_l: bool
    object_storage_use_proxy: bool
    object_storage_set_public_read: bool
    object_storage_s3_force_path_style: bool
    server_rules: NotRequired[
        list[str]
    ]  # v13.11.3以降のバージョンから追加。その場合は使わないとエラー出るかも



from __future__ import annotations

from typing import TYPE_CHECKING
from mipac.actions.admins.accounts import AdminAccountActions

from mipac.actions.admins.user import ClientAdminUserActions, AdminUserActions
from mipac.http import HTTPClient

if TYPE_CHECKING:
    from mipac.client import ClientManager


class ClientAdminUserManager:
    def __init__(self, user_id: str, *, session: HTTPClient, client: ClientManager):
        self.account: AdminAccountActions = AdminAccountActions(session=session, client=client)
        self.__action: ClientAdminUserActions = ClientAdminUserActions(
            user_id=user_id, session=session, client=client
        )

    @property
    def action(self) -> ClientAdminUserActions:
        return self.__action


class AdminUserManager:
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        self.__session: HTTPClient = session
        self.__client: ClientManager = client

    @property
    def action(self) -> AdminUserActions:
        return AdminUserActions(session=self.__session, client=self.__client)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from mipac.config import config
from mipac.types.emoji import ICustomEmojiLite

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class PartialCustomEmoji:
    def __init__(self, emoji: ICustomEmojiLite, *, client: ClientManager) -> None:
        self.__emoji = emoji
        self.__client: ClientManager = client

    @property
    def name(self) -> str:
        return self.__emoji["name"]

    @property
    def url(self) -> str | None:
        protocol = "https" if config.is_ssl else "http"
        url = f"{protocol}://{config.host}/emoji/{self.name}.webp"
        return url

    def _get(self, key: str) -> Any | None:
        return self.__emoji.get(key)



from __future__ import annotations

from typing import TYPE_CHECKING, AsyncGenerator, override

from mipac.abstract.action import AbstractAction
from mipac.http import HTTPClient, Route
from mipac.models.clip import Clip
from mipac.models.note import Note
from mipac.types.clip import IClip
from mipac.types.note import INote
from mipac.utils.pagination import Pagination

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class SharedClipActions(AbstractAction):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        self._session = session
        self._client = client

    async def get_notes(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        *,
        clip_id: str,
    ) -> list[Note]:
        body = {"clipId": clip_id, "limit": limit, "sinceId": since_id, "untilId": until_id}

        raw_notes: list[INote] = await self._session.request(
            Route("POST", "/api/clips/notes"), json=body
        )

        return [Note(raw_note=raw_note, client=self._client) for raw_note in raw_notes]

    async def get_all_notes(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        *,
        clip_id: str,
    ) -> AsyncGenerator[Note, None]:
        """Get notes from a clip

        Parameters
        ----------
        clip_id : str | None, optional, default=None
            The clip id
        limit : int, optional, default=10
            The number of notes to get
        since_id : str | None, optional, default=None
            The note id to get notes after
        until_id : str | None, optional, default=None
            The note id to get notes before

        Yields
        ------
        AsyncGenerator[Note, None]
            The notes
        """
        body = {"clipId": clip_id, "limit": limit, "sinceId": since_id, "untilId": until_id}

        pagination = Pagination[INote](
            self._session, Route("POST", "/api/clips/notes"), json=body, auth=True
        )

        while pagination.is_final is False:
            raw_clips = await pagination.next()
            for raw_clip in raw_clips:
                yield Note(raw_clip, client=self._client)

    async def add_note(self, note_id: str, *, clip_id: str) -> bool:
        """Add a note to a clip

        Parameters
        ----------
        clip_id : str | None, optional, default=None
            The clip id
        note_id : str
            The note id

        Returns
        -------
        bool
            True if the note was added to the clip, False otherwise
        """
        body = {"clipId": clip_id, "noteId": note_id}
        result: bool = await self._session.request(
            Route("POST", "/api/clips/add-note"), json=body, auth=True
        )
        return result

    async def remove_note(self, note_id: str, *, clip_id: str) -> bool:
        """Remove a note from a clip

        Parameters
        ----------
        clip_id : str | None, optional, default=None
            The clip id
        note_id : str
            The note id

        Returns
        -------
        bool
            True if the note was removed from the clip, False otherwise
        """
        body = {"clipId": clip_id, "noteId": note_id}
        result: bool = await self._session.request(
            Route("POST", "/api/clips/remove-note"), json=body, auth=True
        )
        return result

    async def delete(self, *, clip_id: str) -> bool:
        """Delete a clip

        Parameters
        ----------
        clip_id : str | None, optional, default=None
            The clip id

        Returns
        -------
        bool
            True if the clip was deleted, False otherwise
        """
        body = {"clipId": clip_id}
        result: bool = await self._session.request(
            Route("POST", "/api/clips/delete"), json=body, auth=True
        )
        return result

    async def update(
        self,
        name: str,
        is_public: bool | None = None,
        description: str | None = None,
        *,
        clip_id: str,
    ) -> Clip:
        """Update a clip

        Parameters
        ----------
        clip_id : str | None, optional, default=None
            The clip id
        name : str
            The clip name
        is_public : bool, optional
            Whether the clip is public, default=None
        description : str, optional
            The clip description, default=None

        Returns
        -------
        bool
            True if the clip was updated, False otherwise
        """
        body = {"clipId": clip_id, "name": name, "isPublic": is_public, "description": description}
        result: IClip = await self._session.request(
            Route("POST", "/api/clips/update"), json=body, auth=True
        )
        return Clip(result, client=self._client)


class ClientClipActions(SharedClipActions):
    def __init__(self, clip_id: str, *, session: HTTPClient, client: ClientManager):
        super().__init__(session=session, client=client)
        self._clip_id = clip_id

    @override
    async def get_notes(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
    ) -> list[Note]:
        return await super().get_notes(
            limit=limit, since_id=since_id, until_id=until_id, clip_id=self._clip_id
        )

    @override
    async def get_all_notes(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
    ) -> AsyncGenerator[Note, None]:
        """Get notes from a clip

        Parameters
        ----------
        limit : int, optional, default=10
            The number of notes to get
        since_id : str | None, optional, default=None
            The note id to get notes after
        until_id : str | None, optional, default=None
            The note id to get notes before
        get_all : bool, optional, default=False
            Whether to get all notes

        Yields
        ------
        AsyncGenerator[Note, None]
            The notes
        """
        async for note in super().get_all_notes(
            limit=limit, since_id=since_id, until_id=until_id, clip_id=self._clip_id
        ):
            yield note

    @override
    async def add_note(self, note_id: str) -> bool:
        """Add a note to a clip

        Parameters
        ----------
        note_id : str
            The note id

        Returns
        -------
        bool
            True if the note was added to the clip, False otherwise
        """
        return await super().add_note(note_id=note_id, clip_id=self._clip_id)

    @override
    async def remove_note(self, note_id: str) -> bool:
        """Remove a note from a clip

        Parameters
        ----------
        note_id : str
            The note id

        Returns
        -------
        bool
            True if the note was removed from the clip, False otherwise
        """
        return await super().remove_note(note_id=note_id, clip_id=self._clip_id)

    @override
    async def delete(self) -> bool:
        """Delete a clip

        Returns
        -------
        bool
            True if the clip was deleted, False otherwise
        """
        return await super().delete(clip_id=self._clip_id)

    @override
    async def update(
        self,
        name: str,
        is_public: bool | None = None,
        description: str | None = None,
    ) -> Clip:
        """Update a clip

        Parameters
        ----------
        name : str
            The clip name
        is_public : bool, optional
            Whether the clip is public, default=None
        description : str, optional
            The clip description, default=None

        Returns
        -------
        bool
            True if the clip was updated, False otherwise
        """
        return await super().update(
            name=name, is_public=is_public, description=description, clip_id=self._clip_id
        )


class ClipActions(SharedClipActions):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        super().__init__(session=session, client=client)

    async def get_my_favorites(self):
        """Get my favorite clips

        Returns
        -------
        list[Clip]
            The favorite clips
        """
        clips: list[INote] = await self._session.request(
            Route("POST", "/api/clips/my-favorites"), auth=True
        )
        return [Note(raw_clip, client=self._client) for raw_clip in clips]

    async def create(
        self, name: str, is_public: bool = False, description: str | None = None
    ) -> Clip:
        """Create a clip

        Parameters
        ----------
        name : str
            The clip name
        is_public : bool, optional
            Whether the clip is public, default=False
        description : str, optional
            The clip description, default=None

        Returns
        -------
        Clip
            The created clip
        """
        body = {"name": name, "isPublic": is_public, "description": description}
        clip: IClip = await self._session.request(
            Route("POST", "/api/clips/create"), json=body, auth=True
        )
        return Clip(clip, client=self._client)

    async def get_list(self) -> list[Clip]:
        """Get my clips

        Returns
        -------
        list[Clip]
            The clips
        """
        clips: list[IClip] = await self._session.request(
            Route("POST", "/api/clips/list"), auth=True
        )
        return [Clip(raw_clip, client=self._client) for raw_clip in clips]

    async def get(self, clip_id: str) -> Clip:
        """Get a clip

        Parameters
        ----------
        clip_id : str
            The clip id

        Returns
        -------
        Clip
            The clip
        """
        body = {"clipId": clip_id}
        clip: IClip = await self._session.request(
            Route("POST", "/api/clips/show"), json=body, auth=True
        )
        return Clip(clip, client=self._client)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any

from mipac.types.announcement import (
    AnnouncementDisplay,
    AnnoucementIcon,
    IAnnouncement,
    IAnnouncementDetailed,
)
from mipac.utils.format import str_to_datetime

if TYPE_CHECKING:
    from mipac.manager.admins.announcement import ClientAdminAnnouncementManager
    from mipac.manager.client import ClientManager


class Announcement:
    def __init__(self, announcement: IAnnouncement, *, client: ClientManager) -> None:
        self.__announcement: IAnnouncement = announcement
        self.__client: ClientManager = client

    @property
    def id(self) -> str:
        """アナウンスのID

        Returns
        -------
        str
            アナウンスのID
        """
        return self.__announcement["id"]

    @property
    def created_at(self) -> datetime:
        """アナウンスが作成された日時

        Returns
        -------
        datetime
            アナウンスが作成された日時
        """
        return str_to_datetime(self.__announcement["created_at"])

    @property
    def updated_at(self) -> datetime | None:
        """アナウンスが更新された日時

        Noneの場合は、アナウンスが更新されていないことを示します。

        Returns
        -------
        datetime | None
            アナウンスが更新された日時
        """
        return (
            str_to_datetime(self.__announcement["updated_at"])
            if self.__announcement["updated_at"]
            else None
        )

    @property
    def text(self) -> str:
        """アナウンスの内容

        Returns
        -------
        str
            アナウンスの内容
        """
        return self.__announcement["text"]

    @property
    def title(self) -> str:
        """アナウンスのタイトル

        Returns
        -------
        str
            アナウンスのタイトル
        """
        return self.__announcement["title"]

    @property
    def image_url(self) -> str | None:
        return self.__announcement["image_url"]

    @property
    def icon(self) -> AnnoucementIcon:
        return self.__announcement["icon"]

    @property
    def display(self) -> AnnouncementDisplay:
        return self.__announcement["display"]

    @property
    def need_confirmation_to_read(self) -> bool:
        return self.__announcement["need_confirmation_to_read"]

    @property
    def silence(self) -> bool:
        return self.__announcement["silence"]

    @property
    def for_you(self) -> bool:
        return self.__announcement["for_you"]

    @property
    def is_read(self) -> bool | None:
        return self.__announcement.get("is_read")

    def __eq__(self, __value: object) -> bool:
        return isinstance(__value, Announcement) and self.id == __value.id

    def __ne__(self, __value: object) -> bool:
        return not self.__eq__(__value)

    @property
    def api(self) -> ClientAdminAnnouncementManager:
        return self.__client.admin._create_client_announcement_manager(announce_id=self.id)

    def _get(self, key: str) -> Any | None:
        return self.__announcement.get(key)


class AnnouncementDetailed:
    def __init__(self, raw_announcement: IAnnouncementDetailed, *, client: ClientManager) -> None:
        self.__raw_announcement: IAnnouncementDetailed = raw_announcement
        self.__client: ClientManager = client

    @property
    def id(self) -> str:
        return self.__raw_announcement["id"]

    @property
    def created_at(self) -> datetime:
        return str_to_datetime(self.__raw_announcement["created_at"])

    @property
    def updated_at(self) -> datetime | None:
        return (
            str_to_datetime(self.__raw_announcement["updated_at"])
            if self.__raw_announcement["updated_at"]
            else None
        )

    @property
    def text(self) -> str:
        return self.__raw_announcement["text"]

    @property
    def title(self) -> str:
        return self.__raw_announcement["title"]

    @property
    def image_url(self) -> str | None:
        return self.__raw_announcement["image_url"]

    @property
    def icon(self) -> AnnoucementIcon:
        return self.__raw_announcement["icon"]

    @property
    def display(self) -> AnnouncementDisplay:
        return self.__raw_announcement["display"]

    @property
    def is_active(self) -> bool:
        return self.__raw_announcement["is_active"]

    @property
    def for_existing_users(self) -> bool:
        return self.__raw_announcement["for_existing_users"]

    @property
    def silence(self) -> bool:
        return self.__raw_announcement["silence"]

    @property
    def need_confirmation_to_read(self) -> bool:
        return self.__raw_announcement["need_confirmation_to_read"]

    @property
    def user_id(self) -> str | None:
        return self.__raw_announcement["user_id"]

    @property
    def reads(self) -> int:
        """Returns the number of reads of the announcement."""
        return self.__raw_announcement["reads"]

    def __eq__(self, __value: object) -> bool:
        return isinstance(__value, AnnouncementDetailed) and self.id == __value.id

    def __ne__(self, __value: object) -> bool:
        return not self.__eq__(__value)

    @property
    def api(self) -> ClientAdminAnnouncementManager:
        return self.__client.admin._create_client_announcement_manager(announce_id=self.id)

    def _get(self, key: str) -> Any | None:
        return self.__raw_announcement.get(key)



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, TypedDict


if TYPE_CHECKING:
    from mipac.types.user import IUserDetailedNotMeSchema
    from mipac.types.user import IMeDetailedSchema


class IRoleUser(TypedDict):
    id: str
    created_at: str
    user: IUserDetailedNotMeSchema | IMeDetailedSchema
    expires_at: str | None


class IRolePolicies(TypedDict):
    gtl_available: bool
    ltl_available: bool
    can_public_note: bool
    mention_limit: int
    can_invite: bool
    invite_limit: int
    invite_limit_cycle: int
    invite_expiration_time: int
    can_manage_custom_emojis: bool
    can_manage_avatar_decorations: bool
    can_search_notes: bool
    can_use_translator: bool
    can_hide_ads: bool
    drive_capacity_mb: int
    always_mark_nfsw: bool
    pin_limit: int
    antenna_limit: int
    word_mute_limit: int
    webhook_limit: int
    clip_limit: int
    note_each_clips_limit: int
    user_list_limit: int
    user_each_user_lists_limit: int
    rate_limit_factor: int
    avatar_decoration_limit: int


class IPartialRole(TypedDict):
    id: str
    name: str
    color: str | None
    icon_url: str | None
    description: str  # 空でもNoneではない
    is_moderator: bool
    is_administrator: bool
    display_order: int


class IRole(IPartialRole):
    created_at: str
    updated_at: str
    target: Literal["manual", "conditional"]
    cond_formula: dict
    is_public: bool
    is_explorable: bool
    as_badge: bool
    can_edit_members_by_moderator: bool
    policies: IRolePolicies
    users_count: int



from typing import Literal, TypedDict

IAdPlaces = Literal["square" "horizontal" "horizontal-big"]
IAdPriority = Literal["high" "middle" "low"]


class IPartialAd(TypedDict):
    id: str
    url: str
    place: IAdPlaces
    ratio: int
    image_url: str
    day_of_week: int


class IAd(IPartialAd):
    expires_at: str
    starts_at: str
    priority: IAdPriority
    memo: str | None



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.abstract.manager import AbstractManager
from mipac.actions.admins.announcement import (
    AdminAnnouncementActions,
    ClientAdminAnnouncementActions,
)
from mipac.http import HTTPClient

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class ClientAdminAnnouncementManager(AbstractManager):
    def __init__(self, announce_id: str, *, session: HTTPClient, client: ClientManager):
        self.__announce_id: str = announce_id
        self.__session: HTTPClient = session
        self.__client: ClientManager = client

    @property
    def action(self) -> ClientAdminAnnouncementActions:
        return ClientAdminAnnouncementActions(
            announce_id=self.__announce_id, session=self.__session, client=self.__client
        )


class AdminAnnouncementManager(AbstractManager):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        self.__session: HTTPClient = session
        self.__client: ClientManager = client
        self.__action: AdminAnnouncementActions = AdminAnnouncementActions(
            session=self.__session, client=self.__client
        )

    @property
    def action(self) -> AdminAnnouncementActions:
        return self.__action



from __future__ import annotations

from functools import lru_cache, wraps
from typing import Any

DEFAULT_CACHE: dict[str, "LRUCache"] = {}


class Node:
    """キャッシュのノード

    Attributes
    ----------
    key : str
        キャッシュのキー
    value : Any
        キャッシュの値
    next : ListNode
        次のノード
    prev : ListNode
        前のノード
    """

    def __init__(self, key: str, value: Any) -> None:
        self.key: str = key
        self.value: Any = value
        self.next: Node | None = None
        self.prev: Node | None = None


class LRUCache:
    def __init__(self, capacity: int = 100) -> None:
        self.items: dict[str, Node] = {}
        # headとtailはダミーノード、この間にノードを追加していく
        self.head_node = Node("", 0)
        self.tail_node = Node("", 0)
        self.head_node.next = self.tail_node  # 先頭 => 末尾
        self.tail_node.prev = self.head_node  # 先頭 <= 末尾
        self.capacity = capacity

    def _remove_node(self, node: Node):
        if node.prev:  # 削除対象の前にいるnodeのnextを削除対象の次のnodeに変更
            node.prev.next = node.next

        if node.next:  # 削除対象の次のnodeのprevを削除対象の前のnodeに変更
            node.next.prev = node.prev

        del self.items[node.key]

    def _add(self, node: Node):
        prev_node = self.tail_node.prev
        next_node = self.tail_node

        if prev_node:
            prev_node.next = node
            node.prev = prev_node

        if next_node:
            next_node.prev = node
            node.next = next_node

        self.items[node.key] = node

    def get(self, key: str) -> Any:
        if key not in self.items:
            raise KeyError

        node = self.items[key]

        # 一度削除して追加しなおすことで最新のノードとして更新する
        self._remove_node(node)
        self._add(node)

        return node.value

    def put(self, key: str, value: Any):
        if key in self.items:  # 既に同一のキーが存在する場合は再登録するために一度消す
            self._remove_node(self.items[key])

        node = Node(key, value)
        self._add(node)

        if len(self.items) > self.capacity:
            if self.head_node.next:  # 基本存在するはずだけど型的に一応やっとく
                self._remove_node(self.head_node.next)


def cache(group: str = "default", override: bool = False):
    """キャッシュを行います"""

    def decorator(func):
        @wraps(func)
        async def wrapper(self, *args, **kwargs):
            key = cache_key_builder(func, self, *args, **kwargs)
            target_cache = DEFAULT_CACHE.get(group)

            if target_cache is None:
                target_cache = LRUCache()
                DEFAULT_CACHE[group] = target_cache

            try:
                hit_item = target_cache.get(key)
                if hit_item and override is False and kwargs.get("cache_override") is None:
                    return hit_item
            except KeyError:
                res = await func(self, *args, **kwargs)
                target_cache.put(key, res)
                return res

        return wrapper

    return decorator


@lru_cache
def cache_key_builder(func, cls, *args, **kwargs):
    """キャッシュのキーを作成します"""
    ordered_kwargs = sorted(kwargs.items())
    key = (func.__module__ or "") + ".{0}" + f"{cls}" + str(args) + str(ordered_kwargs)
    return key



from __future__ import annotations

from typing import TYPE_CHECKING, AsyncGenerator, override

from mipac.abstract.action import AbstractAction
from mipac.http import HTTPClient, Route
from mipac.models.blocking import Blocking
from mipac.models.user import MeDetailed, UserDetailedNotMe, packed_user
from mipac.types.blocking import IBlocking
from mipac.types.user import IUserDetailed
from mipac.utils.pagination import Pagination

if TYPE_CHECKING:
    from mipac.client import ClientManager


class SharedBlockingActions(AbstractAction):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        self._session: HTTPClient = session
        self._client: ClientManager = client

    async def add(self, *, user_id: str) -> UserDetailedNotMe | MeDetailed:
        """ユーザーをブロックします

        Endpoint: `/api/blocking/create`

        Parameters
        ----------
        user_id : str
            対象のユーザー

        Returns
        -------
        UserDetailedNotMe | MeDetailed
            ブロック対象のユーザー情報
        """
        res: IUserDetailed = await self._session.request(
            Route("POST", "/api/blocking/create"), auth=True, json={"userId": user_id}, lower=True
        )
        return packed_user(res, client=self._client)

    async def remove(self, *, user_id: str) -> UserDetailedNotMe | MeDetailed:
        """ユーザーのブロックを解除します

        Endpoint: `/api/blocking/delete`

        Parameters
        ----------
        user_id : str
            対象のユーザー

        Returns
        -------
        UserDetailedNotMe | MeDetailed
            ブロック解除対象のユーザー情報
        """
        res: IUserDetailed = await self._session.request(
            Route("POST", "/api/blocking/delete"), auth=True, json={"userId": user_id}, lower=True
        )
        return packed_user(res, client=self._client)


class ClientBlockingActions(SharedBlockingActions):
    """クライアント用のブロックアクション

    基本的にoverride以外は行わない
    """

    def __init__(self, user_id: str, *, session: HTTPClient, client: ClientManager):
        super().__init__(session=session, client=client)
        self.__user_id: str = user_id

    @override
    async def add(self) -> UserDetailedNotMe | MeDetailed:
        """ユーザーをブロックします

        Endpoint: `/api/blocking/create`

        Returns
        -------
        UserDetailedNotMe | MeDetailed
            ブロック対象のユーザー情報
        """
        return await super().add(user_id=self.__user_id)

    @override
    async def remove(self) -> UserDetailedNotMe | MeDetailed:
        """ユーザーのブロックを解除します

        Endpoint: `/api/blocking/delete`

        Returns
        -------
        UserDetailedNotMe | MeDetailed
            ブロック解除対象のユーザー情報
        """
        return await super().remove(user_id=self.__user_id)


class BlockingActions(SharedBlockingActions):
    """ブロックアクション
    user_idを持たないメソッドのみを持ち、持つものはSharedBlockingActionsに実装する
    """

    def __init__(self, *, session: HTTPClient, client: ClientManager):
        super().__init__(session=session, client=client)

    async def get_list(
        self,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 100,
    ) -> list[Blocking]:
        """ブロックしているユーザーの一覧を取得します

        Endpoint: `/api/blocking/list`

        Parameters
        ----------
        since_id : str | None, optional
            指定したIDのユーザーより後のユーザーを取得します, default=None
        until_id : str | None, optional
            指定したIDのユーザーより前のユーザーを取得します, default=None
        limit : int, optional
            一度に取得する件数, default=100

        Returns
        -------
        list[Blocking]
            ブロックしているユーザーの一覧
        """
        body = {"limit": limit, "sinceId": since_id, "untilId": until_id}
        raw_blocking_list = await self._session.request(
            Route("POST", "/api/blocking/list"), json=body, auth=True
        )
        return [
            Blocking(raw_blocking=blocking, client=self._client) for blocking in raw_blocking_list
        ]

    async def get_all_list(
        self, limit: int = 30, since_id: str | None = None, until_id: str | None = None
    ) -> AsyncGenerator[Blocking, None]:
        """すべてのブロックしているユーザーの一覧を取得します

        Endpoint: `/api/blocking/list`

        Parameters
        ----------
        limit : int, optional
            一度に取得する件数, default=30
        since_id : str | None, optional
            指定したIDのユーザーより前のユーザーを取得します, default=None
        until_id : str | None, optional
            指定したIDのユーザーより前のユーザーを取得します, default=None

        Yields
        ------
        AsyncGenerator[Blocking, None]
            ブロックしているユーザーの一覧
        """
        pagination = Pagination[IBlocking](
            self._session,
            Route("POST", "/api/blocking/list"),
            json={"limit": limit, "sinceId": since_id, "untilId": until_id},
        )

        while pagination.is_final is False:
            raw_blocking = await pagination.next()
            for raw_blocking in raw_blocking:
                yield Blocking(raw_blocking=raw_blocking, client=self._client)



from __future__ import annotations

import json
import logging
import sys
from typing import Any, Literal
import urllib.parse

import aiohttp

from mipac import __version__
from mipac.config import config
from mipac.errors.base import APIError
from mipac.types.endpoints import ENDPOINTS
from mipac.types.user import IMeDetailedSchema
from mipac.utils.format import remove_dict_empty, upper_to_lower
from mipac.utils.util import COLORS, _from_json, deprecated

_log = logging.getLogger(__name__)


class MisskeyClientWebSocketResponse(aiohttp.ClientWebSocketResponse):
    async def close(self, *, code: int = 4000, message: bytes = b"") -> bool:
        return await super().close(code=code, message=message)


async def json_or_text(response: aiohttp.ClientResponse):
    text = await response.text(encoding="utf-8")
    try:
        if "application/json" in response.headers["Content-Type"]:
            return _from_json(text)
    except KeyError:
        pass


class Route:
    """MisskeyのAPIを使用するためのルートとメソッドを定義するクラス

    Parameters
    ----------
    method : Literal["GET", "POST"]
        メソッド
    path : ENDPOINTS
        ルートパス

    Attributes
    ----------
    path : str
        ルートパス
    method : str
        メソッド
    """

    def __init__(self, method: Literal["GET", "POST"], path: ENDPOINTS):
        self.path: str = path
        self.method: str = method


class HTTPClient:
    """MisskeyのAPIを使用するためのクライアントクラス

    Parameters
    ----------
    url : str
        MisskeyのURL
    token : str | None
        Misskeyのトークン
    """

    def __init__(self, url: str, token: str | None = None) -> None:
        user_agent = (
            "Misskey Bot (https://github.com/yupix/MiPA {0})" + "Python/{1[0]}.{1[1]} aiohttp/{2}"
        )
        parsed_url = urllib.parse.urlparse(url)

        if parsed_url.scheme not in ["http", "https"] or parsed_url.netloc is None:
            raise Exception("Server URL cannot be retrieved or protocol (http / https) is missing")

        protocol = True if parsed_url.scheme == "https" else False

        self._token: str | None = token
        self.user_agent = user_agent.format(__version__, sys.version_info, aiohttp.__version__)
        self._session = aiohttp.ClientSession(ws_response_class=MisskeyClientWebSocketResponse)

        self._url = f"{parsed_url.scheme}://{parsed_url.netloc}"
        config.from_dict(
            host=parsed_url.netloc,
            is_ssl=protocol,
        )

    async def __aenter__(self) -> HTTPClient:
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        await self.close_session()

    @property
    def session(self) -> aiohttp.ClientSession:
        """現在のセッションを返します

        Returns
        -------
        aiohttp.ClientSession
            現在のセッション
        """
        return self._session

    async def request(
        self,
        route: Route,
        auth: bool = True,
        remove_none: bool = True,
        lower: bool = True,
        **kwargs,
    ) -> Any:
        headers: dict[str, str] = {
            "User-Agent": self.user_agent,
        }

        if "json" in kwargs:
            headers["Content-Type"] = "application/json"
            kwargs["json"] = kwargs.pop("json")

        if auth:
            key = "json" if "json" in kwargs or "data" not in kwargs else "data"
            if not kwargs.get(key):
                kwargs[key] = {}
            if self._token:
                kwargs[key]["i"] = self._token

        replace_list = kwargs.pop("replace_list", {})

        for i in ("json", "data"):
            if kwargs.get(i) and remove_none:
                kwargs[i] = remove_dict_empty(kwargs[i])
        async with self._session.request(route.method, self._url + route.path, **kwargs) as res:
            data = await json_or_text(res)
            if lower:
                if isinstance(data, list):
                    data = [upper_to_lower(i, replace_list=replace_list) for i in data]
                if isinstance(data, dict):
                    data = upper_to_lower(data)
            _log.debug(
                f"""{COLORS.green}
REQUEST:{COLORS.reset}
    {kwargs}
{COLORS.green}RESPONSE:{COLORS.reset}
    {json.dumps(data, ensure_ascii=False, indent=4) if data else data}
                """
            )
            if res.status == 204 and data is None:
                return True  # type: ignore
            if 300 > res.status >= 200:
                return data  # type: ignore
            if 511 > res.status >= 300:
                if isinstance(data, dict):
                    APIError(data, res.status).raise_error()
            APIError("HTTP ERROR", res.status).raise_error()

    async def close_session(self) -> None:
        """セッションを終了します"""
        await self._session.close()

    @deprecated
    async def login(self) -> IMeDetailedSchema | None:
        """現在指定されているTokenのユーザー情報を返します

        ..deprecated:: 0.6.4
            :class:`mipac.http.HTTPClient` をインスタンス化した際に自動的にセッションが作成されるようになったため非推奨になりました。現在のTokenのユーザー情報を取得する場合は :meth:`mipac.actions.user.get_me` を使用してください。

        Returns
        -------
        IMeDetailedSchema | None
            Tokenがある場合は自身の情報を返します。Tokenが無い場合はNoneを返します
        """
        if self._token:
            data: IMeDetailedSchema = await self.request(Route("POST", "/api/i"), auth=True)
            return data



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.abstract.action import AbstractAction
from mipac.http import HTTPClient, Route
from mipac.models.app import App
from mipac.types.app import IApp

if TYPE_CHECKING:
    from mipac.client import ClientManager


class AppActions(AbstractAction):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        self.__session: HTTPClient = session
        self.__client: ClientManager = client

    async def create(
        self, name: str, description: str, permission: list[str], callback_url: str | None = None
    ) -> App:
        """Create a new app

        Parameters
        ----------
        name : str
            The name of the app
        description : str
            The description of the app
        permission : list[str]
            The permissions the app has
        callback_url : str, optional
            The callback url of the app, by default None

        Returns
        -------
        App
            The created app
        """

        body = {
            "name": name,
            "description": description,
            "permission": permission,
            "callbackUrl": callback_url,
        }

        raw_app: IApp = await self.__session.request(Route("POST", "/api/app/create"), json=body)
        return App(raw_app)

    async def show(self, app_id: str) -> App:
        """Show an app

        Parameters
        ----------
        app_id : str
            The id of the app

        Returns
        -------
        App
            The app
        """

        raw_app: IApp = await self.__session.request(
            Route("POST", "/api/app/show"), json={"appId": app_id}
        )
        return App(raw_app)



from __future__ import annotations

from typing import TYPE_CHECKING, AsyncGenerator, override

from mipac.abstract.action import AbstractAction
from mipac.http import HTTPClient, Route
from mipac.models.invite import InviteCode, InviteLimit
from mipac.types.invite import IInviteCode, IInviteLimit
from mipac.utils.pagination import Pagination

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class SharedInviteActions(AbstractAction):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        self._session: HTTPClient = session
        self._client: ClientManager = client

    async def delete(self, *, invite_id: str) -> bool:
        """Delete an invite code.

        Parameters
        ----------
        invite_id : str | None, optional
            The invite code to delete, by default None

        Returns
        -------
        bool
            Whether the invite code was deleted.
        """
        res: bool = await self._session.request(
            Route("POST", "/api/invite/delete"), json={"inviteId": invite_id}, auth=True
        )

        return res


class ClientInviteActions(SharedInviteActions):
    def __init__(self, invite_id: str, *, session: HTTPClient, client: ClientManager):
        super().__init__(session=session, client=client)
        self._invite_id: str = invite_id

    @override
    async def delete(self) -> bool:
        """Delete an invite code.

        Returns
        -------
        bool
            Whether the invite code was deleted.
        """
        return await super().delete(invite_id=self._invite_id)


class InviteActions(SharedInviteActions):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        super().__init__(session=session, client=client)

    async def create(self) -> InviteCode:
        """Create a new invite code.

        Endpoint: `/api/invite/create`

        Returns
        -------
        PartialInviteCode
            The invite code created.
        """
        raw_code: IInviteCode = await self._session.request(
            Route("POST", "/api/invite/create"), auth=True
        )
        return InviteCode(raw_code, client=self._client)

    async def get_list(
        self, limit: int = 30, since_id: str | None = None, until_id: str | None = None
    ) -> list[InviteCode]:
        """Get a list of invitation codes created by you.

        Endpoint: `/api/invite/list`

        Parameters
        ----------
        limit : int, optional
            The number of invite codes to get, by default 30
        since_id : str | None, optional
            The id of the invite code to get since, by default None
        until_id : str | None, optional
            The id of the invite code to get until, by default None

        Returns
        -------
        list[PartialInviteCode]
            The list of invite codes.
        """
        data = {"limit": limit, "sinceId": since_id, "untilId": until_id}

        raw_codes: list[IInviteCode] = await self._session.request(
            Route("POST", "/api/invite/list"), auth=True, json=data
        )
        return [InviteCode(raw_code, client=self._client) for raw_code in raw_codes]

    async def get_all_list(
        self, since_id: str | None = None, until_id: str | None = None
    ) -> AsyncGenerator[InviteCode, None]:
        """Get all invite codes created by you.

        Endpoint: `/api/invite/list`

        Parameters
        ----------
        since_id : str | None, optional
            The id of the invite code to get since, by default None
        until_id : str | None, optional
            The id of the invite code to get until, by default None

        Returns
        -------
        list[PartialInviteCode]
            The list of invite codes.
        """
        data = {"limit": 100, "sinceId": since_id, "untilId": until_id}

        pagination = Pagination[IInviteCode](
            self._session, Route("POST", "/api/invite/list"), auth=True, json=data
        )
        while pagination.is_final is False:
            raw_codes: list[IInviteCode] = await pagination.next()
            for raw_code in raw_codes:
                yield InviteCode(raw_code, client=self._client)

    async def get_limit(self) -> InviteLimit:
        """Get the number of invite codes you can create.

        Endpoint: `/api/invite/limit`

        Returns
        -------
        int
            The number of invite codes you can create.
        """
        raw_invite_limit: IInviteLimit = await self._session.request(
            Route("POST", "/api/invite/limit"), auth=True
        )
        return InviteLimit(raw_invite_limit, client=self._client)



from typing import Any
from mipac.types.username import IUsernameAvailable


class UsernameAvailable:
    def __init__(self, raw_username_available: IUsernameAvailable) -> None:
        self.__raw_username_available: IUsernameAvailable = raw_username_available

    @property
    def available(self) -> bool:
        """ユーザー名が使用可能かどうか

        Returns
        -------
        bool
            ユーザー名が使用可能かどうか
        """
        return self.__raw_username_available["available"]

    def _get(self, key: str) -> Any | None:
        return self.__raw_username_available.get(key)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

from mipac.abstract.action import AbstractAction
from mipac.http import HTTPClient, Route
from mipac.models.drive import File
from mipac.types.drive import IFile

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class AdminDriveActions(AbstractAction):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        self.__session: HTTPClient = session
        self.__client: ClientManager = client

    async def clean_remote_files(self) -> bool:
        """Clean remote files

        Endpoint: `/api/admin/drive/clean-remote-files`

        Returns
        -------
        bool
            Whether the remote files were cleaned
        """
        res: bool = await self.__session.request(
            Route("POST", "/api/admin/drive/clean-remote-files"), auth=True
        )
        return res

    async def cleanup(self) -> bool:
        """Clean up the drive

        Endpoint: `/api/admin/drive/cleanup`

        Returns
        -------
        bool
            Whether the drive was cleaned up
        """
        res: bool = await self.__session.request(
            Route("POST", "/api/admin/drive/cleanup"), auth=True
        )
        return res

    async def get_files(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        user_id: str | None = None,
        type: str | None = None,
        origin: Literal["combined", "local", "remote"] = "local",
        hostname: str | None = None,
    ) -> list[File]:
        """Get all files

        Endpoint: `/api/admin/drive/files`

        Parameters
        ----------
        limit: int
            The number of files to get
        since_id: str
            The id of the file to start from
        until_id: str
            The id of the file to end at
        type: str
            The type of file to get
        user_id: str
            The id of the user to get files from
        origin: Literal['combined', 'local', 'remote']
            The origin of the files
        hostname: str
            The hostname of the files

        Returns
        -------
        list[File]
            A list of files
        """
        data = {
            "limit": limit,
            "sinceId": since_id,
            "untilId": until_id,
            "type": type,
            "userId": user_id,
            "origin": origin,
            "hostname": hostname,
        }
        raw_files: list[IFile] = await self.__session.request(
            Route("POST", "/api/admin/drive/files"), auth=True, json=data
        )
        return [File(raw_file=file, client=self.__client) for file in raw_files]

    async def show_file(self, file_id: str, url: str | None = None):
        """Show a file

        Endpoint: `/api/admin/drive/files/show`

        Parameters
        ----------
        file_id: str
            The id of the file to show
        url: str
            The url of the file to show

        Returns
        -------
        dict[str, Any]
            The file
        """
        data = {
            "fileId": file_id,
            "url": url,
        }
        # TODO: IFileではなく、ほぼほぼデータベースの中身が返ってくるのでそれに合わせた型とモデルを作る
        raw_file: dict[str, Any] = await self.__session.request(
            Route("POST", "/api/admin/drive/show-file"), auth=True, json=data
        )
        return raw_file



from mipac.types.app import IApp


class App:
    def __init__(self, raw_app: IApp) -> None:
        self.__raw_app: IApp = raw_app

    @property
    def id(self) -> str:
        """The id of the app"""
        return self.__raw_app["id"]

    @property
    def name(self) -> str:
        """The name of the app"""
        return self.__raw_app["name"]

    @property
    def callback_url(self) -> str | None:
        """The callback url of the app"""
        return self.__raw_app["callback_url"]

    @property
    def permission(self) -> list[str]:
        """The permissions the app has"""
        return self.__raw_app["permission"]

    @property
    def secret(self) -> str:
        """The secret of the app"""
        return self.__raw_app["secret"]

    @property
    def is_authorized(self) -> bool:
        """If the app is authorized or not"""
        return self.__raw_app["is_authorized"]



"""汎用的に使用する関数を定義したモジュール"""

import functools
import json
import warnings
from datetime import datetime, timedelta
from typing import Any


try:
    import orjson  # type: ignore
except ModuleNotFoundError:
    HAS_ORJSON = False
else:
    HAS_ORJSON = True

if HAS_ORJSON:
    _from_json = orjson.loads  # type: ignore
else:
    _from_json = json.loads


class Missing:
    """値が存在しないことを示すクラス"""

    def __repr__(self) -> str:
        return "MISSING"

    def __bool__(self) -> bool:
        return False

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Missing)

    def __ne__(self, other: Any) -> bool:
        return not isinstance(other, Missing)


MISSING: Any = Missing()


class DeprecatedClass:
    def __init__(self, remove_in_version: str) -> None:
        self.remove_in_version = remove_in_version

    def __call__(self, cls):
        remove_in_version = self.remove_in_version

        class Wrapped(cls):
            def __init__(self, *args, **kwargs):
                warnings.simplefilter("always", DeprecationWarning)  # turn off filter
                warnings.warn(
                    "Call to deprecated class {}. (Remove will this class at v{})".format(
                        cls.__name__, remove_in_version
                    ),
                    category=DeprecationWarning,
                    stacklevel=2,
                )
                warnings.simplefilter("default", DeprecationWarning)  # reset filter
                super().__init__(*args, **kwargs)

        return Wrapped


def check_char_length_least(char: str | list[str], length: int) -> bool:
    """指定した文字列の長さが指定した長さ以上かをboolで返します

    Parameters
    ----------
    char : str | list[str]
        確認したい文字列
    length : int
        指定した長さ

    Returns
    -------
    bool
        指定した長さ以上の場合はTrue, それ以外はFalse
    """
    if isinstance(char, str):
        return len(char) >= length
    return all(len(i) >= length for i in char)


def deprecated(func):
    """指定した関数が非推奨であることを示す為のデコレータ"""

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter("always", DeprecationWarning)  # turn off filter
        warnings.warn(
            "Call to deprecated function {}.".format(func.__name__),
            category=DeprecationWarning,
            stacklevel=2,
        )
        warnings.simplefilter("default", DeprecationWarning)  # reset filter
        return func(*args, **kwargs)

    return new_func


def check_multi_arg(*args: Any) -> bool:
    """複数の値を受け取り値が存在するかをboolで返します

    Parameters
    ----------
    args : list
        確認したい変数のリスト

    Returns
    -------
    bool
        存在する場合はTrue, 存在しない場合はFalse
    """
    return bool([i for i in args if i])


class MiTime:
    def __init__(self, start: timedelta, end: datetime):
        self.start = start
        self.end = end


class Colors:
    def __init__(self) -> None:
        self.green = "\x1b[92;1m"
        self.reset = "\x1b[0m"


COLORS = Colors()



from typing import TypedDict


class IHashtag(TypedDict):
    tag: str
    mentioned_users_count: int
    mentioned_local_users_count: int
    mentioned_remote_users_count: int
    attached_users_count: int
    attached_local_users_count: int
    attached_remote_users_count: int


class ITrendHashtag(TypedDict):
    tag: str
    chart: list[int]
    users_count: int



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.abstract.manager import AbstractManager
from mipac.actions.drive.files import ClientFileActions, FileActions
from mipac.http import HTTPClient

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class ClientFileManager(AbstractManager):
    def __init__(self, file_id: str, *, session: HTTPClient, client: ClientManager):
        self.__file_id: str = file_id
        self.__session: HTTPClient = session
        self.__client: ClientManager = client
        self.__action: ClientFileActions = ClientFileActions(
            file_ids=file_id, session=session, client=client
        )

    @property
    def action(self) -> ClientFileActions:
        return self.__action


class DriveFileManager(AbstractManager):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        self.__session: HTTPClient = session
        self.__client: ClientManager = client
        self.__action: FileActions = FileActions(session=session, client=client)

    @property
    def action(self) -> FileActions:
        return self.__action



from __future__ import annotations

from typing import TYPE_CHECKING, override

from mipac.abstract.action import AbstractAction
from mipac.http import HTTPClient, Route
from mipac.models.follow import FollowRequest
from mipac.models.user import PartialUser
from mipac.types.follow import IFollowRequest
from mipac.utils.format import remove_dict_missing
from mipac.utils.util import MISSING, deprecated

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager
    from mipac.types.user import IPartialUser


class SharedFollowActions(AbstractAction):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        self._session = session
        self._client = client

    async def create(self, with_replies: bool = MISSING, *, user_id: str) -> PartialUser:
        data = remove_dict_missing({"userId": user_id, "withReplies": with_replies})

        res: IPartialUser = await self._session.request(
            Route("POST", "/api/following/create"),
            json=data,
        )
        return PartialUser(res, client=self._client)

    @deprecated
    async def add(self, *, user_id: str) -> PartialUser:
        """対象のユーザーをフォローします

        .. deprecated:: 0.6.1
            Use :meth:`mipac.actions.follow.SharedFollowActions.create` instead.

        Returns
        -------
        UserLite:
            The user that you followed
        """
        return await self.create(user_id=user_id)

    async def delete(self, *, user_id: str) -> PartialUser:
        data = {"userId": user_id}
        raw_user: IPartialUser = await self._session.request(
            Route("POST", "/api/following/delete"), json=data, auth=True
        )
        return PartialUser(raw_user=raw_user, client=self._client)

    @deprecated
    async def remove(self, *, user_id: str) -> PartialUser:
        """対象のユーザーのフォローを解除します

        .. deprecated:: 0.6.1
            Use :meth:`mipac.actions.follow.SharedFollowActions.delete` instead.

        Returns
        -------
        PartialUser
            The user that you unfollowed
        """
        return await self.delete(user_id=user_id)

    async def invalidate(self, *, user_id: str) -> PartialUser:
        """
        Make the user unfollows you

        Returns
        -------
        PartialUser
            The user that followed you
        """
        data = {"userId": user_id}
        res: IPartialUser = await self._session.request(
            Route("POST", "/api/following/invalidate"), json=data, auth=True
        )
        return PartialUser(res, client=self._client)


class ClientFollowActions(SharedFollowActions):
    def __init__(self, user_id: str, *, session: HTTPClient, client: ClientManager):
        super().__init__(session=session, client=client)
        self.__user_id: str = user_id

    @override
    async def create(self, with_replies: bool = MISSING) -> PartialUser:
        return await super().create(with_replies, user_id=self.__user_id)

    @override
    async def add(self) -> PartialUser:
        return await super().add(user_id=self.__user_id)

    @override
    async def delete(self) -> PartialUser:
        return await super().delete(user_id=self.__user_id)

    @override
    async def remove(self) -> PartialUser:
        return await super().remove(user_id=self.__user_id)

    @override
    async def invalidate(self) -> PartialUser:
        return await super().invalidate(user_id=self.__user_id)


class FollowActions(SharedFollowActions):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        super().__init__(session=session, client=client)


class FollowRequestActions(AbstractAction):
    def __init__(self, user_id: str | None = None, *, session: HTTPClient, client: ClientManager):
        self.__user_id: str | None = user_id
        self._session = session
        self._client = client

    async def get_all(self) -> list[FollowRequest]:
        """
        Get all follow requests

        Returns
        -------
        list[FollowRequest]
            List of follow requests
        """
        res: list[IFollowRequest] = await self._session.request(
            Route("POST", "/api/following/requests/list"),
            auth=True,
            lower=True,
        )
        return [FollowRequest(follow_request=i, client=self._client) for i in res]

    async def accept(self, user_id: str | None = None) -> bool:
        """
        Accept a follow request

        Parameters
        ----------
        user_id: str
            The user ID to accept

        Returns
        -------
        bool
            Whether the request was accepted
        """
        user_id = user_id or self.__user_id

        data = {"userId": user_id}
        return bool(
            await self._session.request(
                Route("POST", "/api/following/requests/accept"),
                json=data,
                auth=True,
            )
        )

    async def reject(self, user_id: str | None = None) -> bool:
        """
        Reject a follow request

        Parameters
        ----------
        user_id: str
            The user ID to reject

        Returns
        -------
        bool
            Whether the request was rejected
        """
        user_id = user_id or self.__user_id

        data = {"userId": user_id}
        return bool(
            await self._session.request(
                Route("POST", "/api/following/requests/reject"),
                json=data,
                auth=True,
            )
        )

    async def cancel(self, user_id: str | None = None) -> PartialUser:
        """
        Cancel a follow request

        Parameters
        ----------
        user_id: str
            The user ID to cancel

        Returns
        -------
        PartialUser
            The user that you canceled to follow
        """
        user_id = user_id or self.__user_id

        data = {"userId": user_id}
        res: IPartialUser = await self._session.request(
            Route("POST", "/api/following/requests/cancel"),
            json=data,
            auth=True,
            lower=True,
        )
        return PartialUser(res, client=self._client)



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.abstract.manager import AbstractManager
from mipac.actions.admins.emoji import AdminEmojiActions, ClientAdminEmojiActions
from mipac.http import HTTPClient

if TYPE_CHECKING:
    from mipac.client import ClientManager


class ClientAdminEmojiManager(AbstractManager):
    def __init__(self, *, emoji_id: str, session: HTTPClient, client: ClientManager):
        self.__action: ClientAdminEmojiActions = ClientAdminEmojiActions(
            emoji_id=emoji_id, session=session, client=client
        )

    @property
    def action(self) -> ClientAdminEmojiActions:
        return self.__action


class AdminEmojiManager(AbstractManager):
    def __init__(self, session: HTTPClient, client: ClientManager):
        self.__action: AdminEmojiActions = AdminEmojiActions(session=session, client=client)

    @property
    def action(self) -> AdminEmojiActions:
        return self.__action



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any

from mipac.models.lite.user import PartialUser
from mipac.types.invite import IInviteCode, IInviteLimit
from mipac.utils.format import str_to_datetime

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager
    from mipac.manager.invite import ClientInviteManager


class InviteCode:
    def __init__(self, raw_invite_code: IInviteCode, *, client: ClientManager) -> None:
        self.__raw_invite_code: IInviteCode = raw_invite_code
        self.__client = client

    @property
    def id(self) -> str:
        return self.__raw_invite_code["id"]

    @property
    def code(self) -> str:
        return self.__raw_invite_code["code"]

    @property
    def expires_at(self) -> datetime | None:
        return (
            str_to_datetime(self.__raw_invite_code["expires_at"])
            if self.__raw_invite_code["expires_at"]
            else None
        )

    @property
    def created_at(self) -> datetime | None:
        return (
            str_to_datetime(self.__raw_invite_code["created_at"])
            if self.__raw_invite_code["created_at"]
            else None
        )

    @property
    def created_by(self) -> PartialUser | None:
        return (
            PartialUser(self.__raw_invite_code["created_by"], client=self.__client)
            if self.__raw_invite_code["created_by"]
            else None
        )

    @property
    def used_by(self) -> PartialUser | None:
        return (
            PartialUser(self.__raw_invite_code["used_by"], client=self.__client)
            if self.__raw_invite_code["used_by"]
            else None
        )

    @property
    def used_at(self) -> datetime | None:
        return (
            str_to_datetime(self.__raw_invite_code["used_at"])
            if self.__raw_invite_code["used_at"]
            else None
        )

    @property
    def used(self) -> bool:
        return self.__raw_invite_code["used"]

    @property
    def api(self) -> ClientInviteManager:
        return self.__client._create_client_invite_manager(invite_id=self.id)

    def _get(self, key: str) -> Any | None:
        return self.__raw_invite_code.get(key)


class InviteLimit:
    def __init__(self, raw_invite_limit: IInviteLimit, *, client: ClientManager) -> None:
        self.__raw_invite_limit: IInviteLimit = raw_invite_limit
        self.__client = client

    @property
    def remaining(self) -> int | None:
        return self.__raw_invite_limit["remaining"]

    def _get(self, key: str) -> Any | None:
        return self.__raw_invite_limit.get(key)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from mipac.types.hashtag import IHashtag, ITrendHashtag

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class Hashtag:
    def __init__(self, *, raw_hashtag: IHashtag, client: ClientManager) -> None:
        self.__client: ClientManager = client
        self.__raw_hashtag: IHashtag = raw_hashtag

    @property
    def tag(self):
        return self.__raw_hashtag["tag"]

    @property
    def mentioned_users_count(self):
        return self.__raw_hashtag["mentioned_users_count"]

    @property
    def mentioned_local_users_count(self):
        return self.__raw_hashtag["mentioned_local_users_count"]

    @property
    def mentioned_remote_users_count(self):
        return self.__raw_hashtag["mentioned_remote_users_count"]

    @property
    def attached_users_count(self):
        return self.__raw_hashtag["attached_users_count"]

    @property
    def attached_local_users_count(self):
        return self.__raw_hashtag["attached_local_users_count"]

    @property
    def attached_remote_users_count(self):
        return self.__raw_hashtag["attached_remote_users_count"]

    def _get(self, key: str) -> Any | None:
        """生のレスポンスデータに直接アクセスすることができます
        Returns
        -------
        Any | None
            生のレスポンスデータ
        """
        return self.__raw_hashtag.get(key)


class TrendHashtag:
    def __init__(self, *, raw_trend_hashtag: ITrendHashtag, client: ClientManager) -> None:
        self.__raw_trend_hashtag: ITrendHashtag = raw_trend_hashtag
        self.__client: ClientManager = client

    @property
    def tag(self):
        """ハッシュタグ

        Returns
        -------
        str
            ハッシュタグ
        """
        return self.__raw_trend_hashtag["tag"]

    @property
    def chart(self):
        """チャート

        Returns
        -------
        list[int]
            チャート
        """
        return self.__raw_trend_hashtag["chart"]

    @property
    def users_count(self):
        """ユーザー数

        Returns
        -------
        int
            ユーザー数
        """
        return self.__raw_trend_hashtag["users_count"]

    def _get(self, key: str) -> Any | None:
        """生のレスポンスデータに直接アクセスすることができます
        Returns
        -------
        Any | None
            生のレスポンスデータ
        """
        return self.__raw_trend_hashtag.get(key)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any

from mipac.models.note import Note
from mipac.types.channel import IChannel
from mipac.utils.format import str_to_datetime

if TYPE_CHECKING:
    from mipac.manager import ClientManager
    from mipac.manager.channel import ClientChannelManager


class Channel:
    def __init__(self, raw_channel: IChannel, *, client: ClientManager) -> None:
        self._raw_channel: IChannel = raw_channel
        self.__client: ClientManager = client

    @property
    def id(self) -> str:
        """チャンネルID

        Returns
        -------
        str
        """
        return self._raw_channel["id"]

    @property
    def created_at(self) -> datetime:
        """チャンネルが作成された日時

        Returns
        -------
        datetime
        """
        return str_to_datetime(self._raw_channel["created_at"])

    @property
    def last_noted_at(self) -> datetime | None:
        """チャンネルに最後にノートが投稿された日時

        Returns
        -------
        datetime | None
        """
        last_noted_at = self._raw_channel.get("last_noted_at")
        return str_to_datetime(last_noted_at) if last_noted_at else None

    @property
    def name(self) -> str:
        """チャンネル名

        Returns
        -------
        str
        """
        return self._raw_channel["name"]

    @property
    def description(self) -> str | None:
        """チャンネルの説明

        Returns
        -------
        str | None
        """
        return self._raw_channel["description"]

    @property
    def user_id(self) -> str | None:
        """チャンネルを作成したユーザーのID

        Returns
        -------
        str | None
        """
        return self._raw_channel["user_id"]

    @property
    def banner_url(self) -> str | None:
        """チャンネルのバナー画像URL

        Returns
        -------
        str | None
        """
        return self._raw_channel["banner_url"]

    @property
    def pinned_note_ids(self) -> list[str]:
        """ピン留めされているノートのIDのリスト

        Returns
        -------
        list[str]
        """
        return self._raw_channel["pinned_note_ids"]

    @property
    def color(self) -> str:
        """チャンネルの色

        Returns
        -------
        str
        """
        return self._raw_channel["color"]

    @property
    def is_archived(self) -> bool:
        """チャンネルがアーカイブされているかどうか

        Returns
        -------
        bool
        """
        return self._raw_channel["is_archived"]

    @property
    def users_count(self) -> int:
        """チャンネルに参加しているユーザー数

        Returns
        -------
        int
        """
        return self._raw_channel["users_count"]

    @property
    def notes_count(self) -> int:
        """チャンネル内のノート数

        Returns
        -------
        int
        """
        return self._raw_channel["notes_count"]

    @property
    def is_sensitive(self) -> bool:
        """チャンネルがセンシティブかどうか

        Returns
        -------
        bool
        """
        return self._raw_channel["is_sensitive"]

    @property
    def allow_renote_to_external(self) -> bool:
        """外部へのリノートを許可するかどうか

        Returns
        -------
        bool
        """
        return self._raw_channel["allow_renote_to_external"]

    @property
    def is_following(self) -> bool | None:
        """自身がフォローしているかどうか

        Returns
        -------
        bool
        """
        return self._raw_channel.get("is_following")

    @property
    def is_favorited(self) -> bool | None:
        """自身がお気に入り登録しているかどうか

        Returns
        -------
        bool
        """
        return self._raw_channel.get("is_favorited")

    @property
    def pinned_notes(self) -> list[Note]:
        """ピン留めされているノートのリスト

        Returns
        -------
        list[Note]
        """
        return [
            Note(note, client=self.__client) for note in self._raw_channel.get("pinned_notes", [])
        ]

    @property
    def api(self) -> ClientChannelManager:
        """チャンネルに関するAPIを利用するためのクライアント

        Returns
        -------
        ClientChannelManager
        """
        return self.__client._create_client_channel_manager(channel_id=self.id)

    def _get(self, key: str) -> Any | None:
        return self._raw_channel.get(key)

    def __eq__(self, __value: object) -> bool:
        return isinstance(__value, Channel) and self.id == __value.id

    def __ne__(self, __value: object) -> bool:
        return not self.__eq__(__value)



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from mipac.abstract.action import AbstractAction
from mipac.http import HTTPClient, Route
from mipac.models.hashtag import Hashtag, TrendHashtag
from mipac.models.user import MeDetailed, UserDetailedNotMe, packed_user
from mipac.types.hashtag import IHashtag, ITrendHashtag
from mipac.types.user import IMeDetailedSchema, IUserDetailedNotMeSchema

if TYPE_CHECKING:
    from mipac.client import ClientManager


class HashtagActions(AbstractAction):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        self.__session: HTTPClient = session
        self.__client: ClientManager = client

    async def get_list(
        self,
        sort: Literal[
            "+mentionedUsers",
            "-mentionedUsers",
            "+mentionedLocalUsers",
            "-mentionedLocalUsers",
            "+mentionedRemoteUsers",
            "-mentionedRemoteUsers",
            "+attachedUsers",
            "-attachedUsers",
            "+attachedLocalUsers",
            "-attachedLocalUsers",
            "+attachedRemoteUsers",
            "-attachedRemoteUsers",
        ],
        limit: int = 10,
        attached_to_user_only: bool = False,
        attached_to_local_user_only: bool = False,
        attached_to_remote_user_only: bool = False,
    ) -> list[Hashtag]:
        """ハッシュタグのリストを取得します。

        Parameters
        ----------
        sort: Literal[
            "+mentionedUsers",
            "-mentionedUsers",
            "+mentionedLocalUsers",
            "-mentionedLocalUsers",
            "+mentionedRemoteUsers",
            "-mentionedRemoteUsers",
            "+attachedUsers",
            "-attachedUsers",
            "+attachedLocalUsers",
            "-attachedLocalUsers",
            "+attachedRemoteUsers",
            "-attachedRemoteUsers",
        ]
            ソートの方法を指定します。
        limit: int, optional
            取得するハッシュタグの数を指定します, default=10
        attached_to_user_only: bool, optional
            ユーザーに添付されたハッシュタグのみを取得するかどうかを指定します, default=False
        attached_to_local_user_only: bool, optional
            ローカルユーザーに添付されたハッシュタグのみを取得するかどうかを指定します, default=False
        attached_to_remote_user_only: bool, optional
            リモートユーザーに添付されたハッシュタグのみを取得するかどうかを指定します, default=False

        Returns
        -------
        list[Hashtag]
            取得したハッシュタグのリストです。
        """
        body = {
            "limit": limit,
            "attachedToUserOnly": attached_to_user_only,
            "attachedToLocalUserOnly": attached_to_local_user_only,
            "attachedToRemoteUserOnly": attached_to_remote_user_only,
            "sort": sort,
        }

        raw_hashtags: list[IHashtag] = await self.__session.request(
            Route("POST", "/api/hashtags/list"), json=body
        )

        return [
            Hashtag(raw_hashtag=raw_hashtag, client=self.__client) for raw_hashtag in raw_hashtags
        ]

    async def search(self, query: str, limit: int = 10, offset: int = 0) -> list[str]:
        """ハッシュタグを検索します

        Parameters
        ----------
        query: str
            検索するクエリを指定します。
        limit: int, optional
            取得するハッシュタグの数を指定します, default=10
        offset: int, optional
            オフセットを指定します, default=0

        Returns
        -------
        list[Hashtag]
            取得したハッシュタグのリストです。
        """

        body = {"query": query, "limit": limit, "offset": offset}

        raw_hashtags: list[str] = await self.__session.request(
            Route("POST", "/api/hashtags/search"),
            json=body,
            lower=False,  # 戻り値がarrayのstrなのでlowerするとエラーになる
        )

        return raw_hashtags

    async def show(self, tag: str):
        """ハッシュタグの情報を取得します。

        Parameters
        ----------
        tag: str
            取得するハッシュタグの名前です。

        Returns
        -------
        Hashtag
            取得したハッシュタグの情報です。
        """
        body = {"tag": tag}

        raw_hashtag: IHashtag = await self.__session.request(
            Route("POST", "/api/hashtags/show"), json=body
        )

        return Hashtag(raw_hashtag=raw_hashtag, client=self.__client)

    async def get_trend(self):
        """トレンドのハッシュタグを取得します。

        Returns
        -------
        list[TrendHashtag]
            取得したハッシュタグのリストです。
        """
        raw_trend_hashtags: list[ITrendHashtag] = await self.__session.request(
            Route("GET", "/api/hashtags/trend")
        )

        return [
            TrendHashtag(raw_trend_hashtag=raw_trend_hashtag, client=self.__client)
            for raw_trend_hashtag in raw_trend_hashtags
        ]

    async def get_users(
        self,
        tag: str,
        sort: Literal[
            "+follower", "-follower", "+createdAt", "-createdAt", "+updatedAt", "-updatedAt"
        ],
        state: Literal["all", "alive"] = "all",
        origin: Literal["combined", "local", "remote"] = "local",
    ) -> list[UserDetailedNotMe | MeDetailed]:
        body = {"tag": tag, "sort": sort, "state": state, "origin": origin}

        raw_users: list[
            IUserDetailedNotMeSchema | IMeDetailedSchema
        ] = await self.__session.request(Route("POST", "/api/hashtags/users"), json=body)

        return [packed_user(raw_user, client=self.__client) for raw_user in raw_users]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from mipac.actions.admins.user import SharedAdminUserActions
from mipac.http import HTTPClient, Route
from mipac.models.admin import IndexStat, ModerationLog, ServerInfo, UserIP
from mipac.models.meta import AdminMeta
from mipac.types.admin import IIndexStat, IModerationLog, IServerInfo, ITableStats, IUserIP
from mipac.types.meta import IAdminMeta, IUpdateMetaBody
from mipac.utils.cache import cache
from mipac.utils.format import convert_dict_keys_to_camel
from mipac.utils.pagination import Pagination

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class AdminActions(SharedAdminUserActions):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        super().__init__(session=session, client=client)

    async def get_meta(self, detail: bool = False) -> AdminMeta:
        """
        Get admin meta
        Endpoint: `/api/admin/meta`

        Parameters
        ----------
        detail : bool, optional
            flag of detail, by default False
        """
        res: IAdminMeta = await self._session.request(
            Route("POST", "/api/admin/meta"),
            json={"detail": detail},  # 現状detailがあってもなんも変わらない
            auth=True,
            lower=True,
        )
        return AdminMeta(res, client=self._client)

    async def update_user_note(self, user_id: str, text: str) -> bool:
        """
        Update user note
        Endpoint: `/api/admin/update-user-note`

        Parameters
        ----------
        user_id : str
            target user's id
        text : str
            new note
        """
        body = {"userId": user_id, "text": text}
        return bool(
            await self._session.request(
                Route("POST", "/api/admin/update-user-note"), auth=True, json=body
            )
        )

    async def update_meta(self, meta: IUpdateMetaBody) -> bool:  # TODO: 引数を見直すべき
        body = convert_dict_keys_to_camel(
            meta, replace_list={"tos_text_url": "ToSTextUrl", "tos_url": "ToSUrl"}
        )

        return bool(
            await self._session.request(
                Route("POST", "/api/admin/update-meta"),
                json=body,
                auth=True,
                lower=True,
                remove_none=False,
            )
        )

    async def get_moderation_logs(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        type: str | None = None,
        user_id: str | None = None,
    ) -> list[ModerationLog]:
        body = {
            "limit": limit,
            "sinceId": since_id,
            "untilId": until_id,
            "type": type,
            "userId": user_id,
        }

        res_moderation_logs: list[IModerationLog] = await self._session.request(
            Route("POST", "/api/admin/show-moderation-logs"), json=body, auth=True
        )
        return [ModerationLog(res, client=self._client) for res in res_moderation_logs]

    async def get_all_moderation_logs(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        type: str | None = None,
        user_id: str | None = None,
    ):
        body = {
            "limit": limit,
            "sinceId": since_id,
            "untilId": until_id,
            "type": type,
            "userId": user_id,
        }

        pagination = Pagination[IModerationLog](
            self._session, Route("POST", "/api/admin/show-moderation-logs"), json=body
        )

        while pagination.is_final is False:
            res_moderation_logs: list[IModerationLog] = await pagination.next()
            for res_moderation_log in res_moderation_logs:
                yield ModerationLog(res_moderation_log, client=self._client)

    @cache("server_info")
    async def get_server_info(self, **kwargs) -> ServerInfo:
        server_info_payload: IServerInfo = await self._session.request(
            Route("POST", "/api/admin/server-info"), auth=True, lower=True
        )
        return ServerInfo(server_info_payload)

    async def fetch_server_info(self) -> ServerInfo:
        return await self.get_server_info(cache_override=True)

    async def send_email(self, to: str, subject: str, text: str) -> bool:
        """
        Send email to specified address
        Endpoint: `/api/admin/send-email`

        Parameters
        ----------
        to : str
            email address to send
        subject : str
            email subject
        text : str
            email body
        """
        body = {"to": to, "subject": subject, "text": text}
        return bool(
            await self._session.request(
                Route("POST", "/api/admin/send-email"), auth=True, json=body
            )
        )

    async def resolve_abuse_user_report(self, report_id: str, forward: bool = False) -> bool:
        """
        Resolve abuse user report
        Endpoint: `/api/admin/resolve-abuse-user-report`

        Parameters
        ----------
        report_id : str
            report id
        forward : bool, optional
            Whether to forward the report to a remote server
        """
        body = {"reportId": report_id, "forward": forward}
        return bool(
            await self._session.request(
                Route("POST", "/api/admin/resolve-abuse-user-report"), auth=True, json=body
            )
        )

    async def reset_password(self, user_id: str) -> str:
        """
        target user's password reset
        Endpoint: `/api/admin/reset-password`

        Parameters
        ----------
        user_id : str
            target user's id

        Returns
        -------
        str
            new password
        """
        return await self._session.request(
            Route("POST", "/api/admin/reset-password"), auth=True, json={"userId": user_id}
        )

    async def get_table_stats(self) -> dict[str, ITableStats]:  # TODO: モデルにしてもいいかも
        return await self._session.request(Route("POST", "/api/admin/get-table-stats"), auth=True)

    async def get_index_stats(self) -> list[IndexStat]:  # TODO: モデルを改めて確認する
        res: list[IIndexStat] = await self._session.request(
            Route("POST", "/api/admin/get-index-stats"), auth=True
        )
        return [IndexStat(i) for i in res]

    async def get_user_ips(self, user_id: str) -> list[UserIP]:
        res: list[IUserIP] = await self._session.request(
            Route("POST", "/api/admin/get-user-ips"),
            auth=True,
            json={"userId": user_id},
            lower=True,
        )
        return [UserIP(i) for i in res]

    async def show_users(
        self,
        limit: int = 10,
        offset: int = 0,
        sort: str | None = None,
        state: str = "all",
        origin: str = "combined",
        username: str | None = None,
        hostname: str | None = None,
    ):
        body = {
            "limit": limit,
            "offset": offset,
            "sort": sort,
            "state": state,
            "origin": origin,
            "username": username,
            "hostname": hostname,
        }
        res: Any = await self._session.request(
            Route("POST", "/api/admin/show-users"), auth=True, json=body
        )

        return res  # TODO: 専用のモデルを作るべき



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, override

from mipac.http import HTTPClient, Route
from mipac.models.roles import Role, RoleUser
from mipac.types.meta import IPolicies
from mipac.types.roles import IRole, IRoleUser
from mipac.utils.format import remove_dict_missing
from mipac.utils.pagination import Pagination
from mipac.utils.util import MISSING

if TYPE_CHECKING:
    from mipac.http import HTTPClient
    from mipac.manager.client import ClientManager


class SharedAdminRoleActions:
    def __init__(self, *, session: HTTPClient, client: ClientManager) -> None:
        self._session: HTTPClient = session
        self._client: ClientManager = client

    async def delete(self, *, role_id: str) -> bool:
        res: bool = await self._session.request(
            Route("POST", "/api/admin/roles/delete"),
            json={"roleId": role_id},
        )
        return res

    async def update(
        self,
        name: str,
        description: str,
        color: str | None = None,
        iconUrl: str | None = None,
        target: Literal["manual", "conditional"] = "manual",
        cond_formula: dict[Any, Any] | None = None,
        is_public: bool = False,
        is_moderator: bool = False,
        is_administrator: bool = False,
        is_explorable: bool = MISSING,
        as_badge: bool = False,
        can_edit_members_by_moderator: bool = False,
        display_order: int = 0,
        policies: dict[Any, Any] | None = None,
        *,
        role_id: str,
    ) -> bool:
        body = remove_dict_missing(
            {
                "roleId": role_id,
                "name": name,
                "description": description,
                "color": color,
                "iconUrl": iconUrl,
                "target": target,
                "condFormula": cond_formula or {},
                "isPublic": is_public,
                "isModerator": is_moderator,
                "isAdministrator": is_administrator,
                "isExplorable": is_explorable,
                "asBadge": as_badge,
                "canEditMembersByModerator": can_edit_members_by_moderator,
                "displayOrder": display_order,
                "policies": policies or {},
            }
        )
        res: bool = await self._session.request(
            Route("POST", "/api/admin/roles/update"),
            json=body,
            remove_none=False,
        )
        return res

    async def assign(self, user_id: str, expires_at: int | None = None, *, role_id: str) -> bool:
        """指定したユーザーに指定したロールを付与します

        Parameters
        ----------
        user_id : str
            ロールを付与する対象のユーザーID
        expires_at : int | None, optional
            いつまでロールを付与するか, by default None
        role_id : str
            ロールのID

        Returns
        -------
        bool
            成功したか否か
        """
        body = {"roleId": role_id, "userId": user_id, "expiresAt": expires_at}
        res: bool = await self._session.request(
            Route("POST", "/api/admin/roles/assign"), auth=True, json=body
        )
        return res

    async def unassign(self, user_id: str, *, role_id: str) -> bool:
        """指定したユーザーから指定したロールを解除します

        Parameters
        ----------
        user_id : str
            ロールを解除するユーザーID
        role_id : str
            ロールのID

        Returns
        -------
        bool
            成功したか否か
        """
        body = {"roleId": role_id, "userId": user_id}
        res: bool = await self._session.request(
            Route("POST", "/api/admin/roles/unassign"), auth=True, json=body
        )
        return res

    async def get_users(
        self,
        since_id: str = MISSING,
        until_id: str = MISSING,
        limit: int = MISSING,
        *,
        role_id: str,
    ) -> RoleUser:
        body = remove_dict_missing(
            {"limit": limit, "sinceId": since_id, "untilId": until_id, "roleId": role_id}
        )

        raw_role_user: IRoleUser = await self._session.request(
            Route("POST", "/api/admin/roles/users"), json=body
        )

        return RoleUser(raw_role_user, client=self._client)

    async def get_all_users(
        self,
        since_id: str = MISSING,
        until_id: str = MISSING,
        limit: int = MISSING,
        *,
        role_id: str,
    ):
        """指定したロールを持つユーザーを全て取得します

        Parameters
        ----------
        since_id : str, optional
            ページネーションの開始位置, by default MISSING
        until_id : str, optional
            ページネーションの終了位置, by default MISSING
        limit : int, optional
            1ページあたりの取得数, by default MISSING
        role_id : str
            ロールのID

        Returns
        -------
        RoleUser
            ロールユーザー
        """
        body = remove_dict_missing(
            {"sinceId": since_id, "untilId": until_id, "limit": limit, "roleId": role_id}
        )

        pagination = Pagination[IRoleUser](
            self._session, Route("POST", "/api/admin/roles/users"), json=body
        )

        while pagination.is_final is False:
            raw_role_users = await pagination.next()
            for raw_role_user in raw_role_users:
                yield RoleUser(raw_role_user, client=self._client)


class ClientAdminRoleActions(SharedAdminRoleActions):
    def __init__(self, role_id: str, *, session: HTTPClient, client: ClientManager) -> None:
        super().__init__(session=session, client=client)
        self.__role_id: str = role_id

    @override
    async def delete(self) -> bool:
        return await super().delete(role_id=self.__role_id)

    @override
    async def update(
        self,
        name: str,
        description: str,
        color: str | None = None,
        iconUrl: str | None = None,
        target: Literal["manual"] | Literal["conditional"] = "manual",
        cond_formula: dict[Any, Any] | None = None,
        is_public: bool = False,
        is_moderator: bool = False,
        is_administrator: bool = False,
        is_explorable: bool = MISSING,
        as_badge: bool = False,
        can_edit_members_by_moderator: bool = False,
        display_order: int = 0,
        policies: dict[Any, Any] | None = None,
    ) -> bool:
        return await super().update(
            name=name,
            description=description,
            color=color,
            iconUrl=iconUrl,
            target=target,
            cond_formula=cond_formula,
            is_public=is_public,
            is_moderator=is_moderator,
            is_administrator=is_administrator,
            is_explorable=is_explorable,
            as_badge=as_badge,
            can_edit_members_by_moderator=can_edit_members_by_moderator,
            display_order=display_order,
            policies=policies,
            role_id=self.__role_id,
        )

    @override
    async def assign(self, user_id: str, expires_at: int | None = None) -> bool:
        return await super().assign(user_id=user_id, expires_at=expires_at, role_id=self.__role_id)

    @override
    async def unassign(self, user_id: str) -> bool:
        return await super().unassign(user_id=user_id, role_id=self.__role_id)

    @override
    async def get_users(
        self,
        since_id: str = MISSING,
        until_id: str = MISSING,
        limit: int = MISSING,
    ) -> RoleUser:
        return await super().get_users(
            since_id=since_id, until_id=until_id, limit=limit, role_id=self.__role_id
        )

    @override
    async def get_all_users(
        self,
        since_id: str = MISSING,
        until_id: str = MISSING,
        limit: int = MISSING,
    ):
        return super().get_all_users(
            since_id=since_id, until_id=until_id, limit=limit, role_id=self.__role_id
        )


class AdminRoleActions(SharedAdminRoleActions):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        super().__init__(session=session, client=client)

    async def create(
        self,
        name: str,
        description: str,
        color: str | None = None,
        iconUrl: str | None = None,
        target: Literal["manual", "conditional"] = "manual",
        cond_formula: dict[Any, Any] | None = None,
        is_public: bool = False,
        is_moderator: bool = False,
        is_administrator: bool = False,
        is_explorable: bool = False,
        as_badge: bool = False,
        can_edit_members_by_moderator: bool = False,
        display_order: int = 0,
        policies: dict[Any, Any] | None = None,
    ) -> Role:
        body = {
            "name": name,
            "description": description,
            "color": color,
            "iconUrl": iconUrl,
            "target": target,
            "condFormula": cond_formula or {},
            "isPublic": is_public,
            "isModerator": is_moderator,
            "isAdministrator": is_administrator,
            "isExplorable": is_explorable,
            "asBadge": as_badge,
            "canEditMembersByModerator": can_edit_members_by_moderator,
            "displayOrder": display_order,
            "policies": policies or {},
        }
        res: IRole = await self._session.request(
            Route("POST", "/api/admin/roles/create"),
            auth=True,
            json=body,
            lower=True,
            remove_none=False,
        )
        return Role(res, client=self._client)

    async def get_list(self) -> list[Role]:
        res: list[IRole] = await self._session.request(
            Route("POST", "/api/admin/roles/list"), auth=True, lower=True
        )
        return [Role(i, client=self._client) for i in res]

    async def update_default_policies(self, policies: IPolicies):
        body = {
            "policies": {
                "gtlAvailable": policies.get("gtl_available"),
                "ltlAvailable": policies.get("ltl_available"),
                "canPublicNote": policies.get("can_public_note"),
                "canInvite": policies.get("can_invite"),
                "canManageCustomEmojis": policies.get("can_manage_custom_emojis"),
                "canHideAds": policies.get("can_hide_ads"),
                "driveCapacityMb": policies.get("drive_capacity_mb"),
                "pinLimit": policies.get("pin_limit"),
                "antennaLimit": policies.get("antenna_limit"),
                "wordMuteLimit": policies.get("word_mute_limit"),
                "webhookLimit": policies.get("webhook_limit"),
                "clipLimit": policies.get("clip_limit"),
                "noteEachClipsLimit": policies.get("note_each_clips_limit"),
                "userListLimit": policies.get("user_list_limit"),
                "userEachUserListsLimit": policies.get("user_each_user_lists_limit"),
                "rateLimitFactor": policies.get("rate_limit_factor"),
            }
        }
        res = await self._session.request(
            Route("POST", "/api/admin/roles/update-default-policies"),
            auth=True,
            lower=True,
            json=body,
        )
        return res
